// Sequencer v1.12

// An applet to maintain a set of number sequences and to enable references to entries
// in the sequence.
//
// Written by Roger Noble, June 2000

// Version History:
// v1.0  (1st June 2000):  Initial version.
// V1.01 (6th June 2000):  Adjusted maximum anchor number if last anchor in a sequence deleted.
//                         Fixed reference searching fault when a reference is deleted.
// v1.02 (11th June 2000): Adjusted maximum anchor number when sequence re-ordered.
//                         When an anchor was deleted, the incorrect next or previous anchor
//                         was being selected for display.
// v1.03 (12th June 2000): Updated control window correctly when an anchor or reference is deleted.
// v1.04 (30th Nov 2000):  Significant update in speed when renumbering or reordering. The problem
//                         here was that when replacing an existing anchor or reference text we were
//                         selecting it and then overtyping it with the new value. However, because
//                         of a bug in Rupert Thompson's enhanced bookmark manager this takes ages
//			   when there are as lot of bookmarks, so I changed the overtype with
//                         the appropriate number of deletions followed by inserting the replacement
//                         text. Actually, I fixed the enhanced bookmark manager, but this raised
//                         another issue: this mananager would delete the bookmarks in the referenced
//                         text when it was overwritten, but the ordinary bookmark manager would not.
//                         Since I need to know whether the bookmarks were being deleted or not, it
//                         was simplest to use the deletions method anyway.
// v1.05 (6th April 2001): Display hour glass when busy.
// v1.10 (13th July 2001): Added push button to delete the current sequence.
//                         Added support for !!!!!!!bm, so that we tidy ourselves up if any of
//                         our bookmarks are deleted by the user.
// v1.11 (4th Aug 2001):   Removed delete word etc keypress handling. 7!bm does this for us by
//                         setting the appropriate zone and calling our selection deletion
//                         handler.
// v1.12 (21st May 2002):  bmcmp() does not appear to return the correct value for bookmarks in
//                         separate chapters, so added a separate check for the chapter number
//                         in Seq_set_order().

//=============================================================================================

// Terminology:

// Sequence    A linear sequence of integers in increasing numerical order; consecutive entries in
//             the sequence differ by one.
// Anchor      A value from a sequence that is inserted into a document. For example, a figure number.
// Reference   A reference to an anchor inserted into a document. If the value of the anchor is
//             changed then the value of the reference changes also. For example, a reference to
//             a certain figure.

//=============================================================================================

// Important Notes.

// Sequencer saves the parameters of the sequencies that it maintains as tags in the Ovation file,
// and anchors and references are indicated by bookmarks. As of version 2.58 of Ovation Pro tags
// and bookmarks are not saved if the document is stored as a DDL file. So if you save your
// document as DDL then any documents re-created from that file will not contain the sequence
// information. The inserted anchors and references will still be there, they will just not be
// recognised as belonging to a sequence.

// It is possible to accidently delete a bookmark required by Sequencer. Bookmarks are placed
// immediately before anchors and references. Always use the Delete button in the Sequencer
// control window to remove anchors and references, and be careful when deleting any text
// just before an anchor or reference. If you inadvertently delete a bookmark you can insert
// it back by hand so long as you put it in the right place and give it the right name (see
// below). Alternatively, if you renumber or reorder the sequence then Sequencer will remove
// entries with non-existant bookmarks from the tags stored with the document. You can then add
// a new entry as required.

//=============================================================================================

// Technical Notes.

// Anchor bookmarks are all of the form SeqA_<seq_no>_<anchor_no> where <seq_no> is the
// number of the sequence (the first sequence is numbered zero), and <anchor_no> is the
// number of the anchor (the first anchor in a sequence is numbered zero).
// Reference boomarks are of the form SeqR_<seq_no>_<anchor_no>_<ref_no> where <ref_no>
// is the number of the reference (the first reference to an anchor is numbered zero).

// Sequencer uses tags as follows, all of the tag numbers given are relative to the base
// tag value used for Sequencer.

// Tag       Index               Contents
//
// 0         0                   Number of sequencies
// 1         0                   Version number * 100

// These are followed by three tags plus index values for each sequence. To get the
// tag offset for a given sequence then use <seq_no> * 3 + 2. The tag number below
// is relative to the start of the tags for a sequence.

// 0         -                   Sequence name (string)
// 1         0                   Sequence starting number
// 1         1                   Maximum anchor number in sequence
// 1         2                   Last requested number format (see below)
// 1         3                   Number of anchors in the sequence (0 if sequence unused)
// 1         4                   Current number format for the sequence (see below)

// 2         0                   Anchor value for first anchor in sequence
// 2         1                   Number of references to first anchor (-1 if anchor deleted)
// 2         2                   Ordering number
// 2         3                   Maximum reference number
// 2         4                   Anchor value for second anchor in sequence
// 2         5                   Number of references to second anchor (-1 if anchor deleted)
// 2         6                   Ordering number
// 2         7                   Maximum reference number
// etc       etc                 etc

// Number formats are as follows:
// 2 = arabic
// 1 = roman lower case
// 0 = roman upper case

// The ordering number indicates the order of the anchor in the document. Anchors with
// Low ordering numbers are located before high ordering numbers in the document.

// As implied above, once created a sequence is never deleted, but it is marked as being free
// if it contains no anchors.
// Similarly, an anchor value entry in the tags is never removed, but it can be marked as being
// free by setting its reference count to -1.
// These are done to ensure that the anchor and reference numbers stored in the embedded bookmarks
// do not need to be changed if an anchor or reference is deleted.

//=============================================================================================

// Global Variables.

// Constants.

int Seq_Version = 103;                            // Current version number * 100

int Seq_tag = ('s'<<24) | ('e'<<16) | ('q'<<8);   // Base tag value
int Seq_max_seq = 32;			          // Maximum number of sequences allowed (each sequence uses three tags)
                                                  // The maximum number of tags allocated to any one applet should not
                                                  // be greater then 256.
int Seq_seq_tags = 3;                             // Number of tags stored per sequence
int Seq_anchor_tags = 4;                          // Number of index values per anchor tag 2

// Global sequence stuff.

int Seq_number_of_seqs = 0;                       // Number of active sequences

// Options.

int Seq_NumberFormat = 2;  // Number format: 0 = Roman upper, 1 = Roman lower,
                           // 2 = Arabic
int Seq_Start = 1;         // Starting number
string Seq_Name;           // Sequence name

// Handles and techie things.

int Seq_win;           // The control window handle
int Seq_menu = 0;      // Menu handle for entries in sequence

// Current values.

int Seq_current_seq = -1;         // Current sequence
int Seq_current_entry = -1;       // Current entry in the sequence
int Seq_current_entries = 0;      // Number of entries in the sequence
int Seq_max_entry_value = -1;     // Current maximum entry value in sequence
int Seq_current_entry_value = -1; // Current entry value
int Seq_current_references = 0;   // Number of references to current anchor
int Seq_max_reference = 0;        // Maximum reference number for current anchor
int Seq_current_ref = 0;          // Used to keep track of which reference we are currently one when searching
int Seq_CurrentNumberFormat = -1; // Existing number format

// Document-related stuff.

int Seq_current_file = 0; 			       // Handle to current file
int Seq_current_view = 0;			       // Handle to current view


//=============================================================================================

////////////////////////////////////////////////
// Sequence handling functions                //
////////////////////////////////////////////////

// Finds the previous undeleted sequence entry. Returns -1 if none.
// Previous here refers to the previous ordering entry, and so to
// the previous anchor as they are placed in the document.

int Seq_previous_entry(int entry)
{
  int offset = Seq_current_seq * Seq_seq_tags + 2;
  int location;
  int i;

  // Obtain the ordering number of the current entry.

  if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+2))
    location = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+2);
  else
    location = 0;

  // Search for the closest previous anchor that is not deleted.

  entry = -1;
  while (entry < 0 && --location >= 0)
  {
    // Search for the previous ordering number.

    for (i = 0; i < Seq_current_entries; ++i)
      if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1))
        if (tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1) >= 0)  // Not deleted
          if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+2))
            if (tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+2) == location)
            {
              entry = i;
              break;
            }
  }

  return entry;
}

//---------------------------------------------------------------------------------------------

// Finds the next undeleted sequence entry. Returns -1 if none.
// Next here refers to the next ordering entry, and so to
// the next anchor as they are placed in the document.

int Seq_next_entry(int entry)
{
  int offset = Seq_current_seq * Seq_seq_tags + 2;
  int location;
  int i;

  // Obtain the ordering number of the current entry.

  if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+2))
    location = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+2);
  else
    location = Seq_current_entries;

  // Search for the closest next anchor that is not deleted.

  entry = -1;
  while (entry < 0 && ++location < Seq_current_entries)
  {
    // Search for the next ordering number.

    for (i = 0; i < Seq_current_entries; ++i)
      if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1))
        if (tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1) >= 0)  // Not deleted
          if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+2))
            if (tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+2) == location)
            {
              entry = i;
              break;
            }
  }

  return entry;
}

//---------------------------------------------------------------------------------------------

// Finds the first sequence entry as they are ordered in the document.
// Returns -1 if there are no entries.

int Seq_first_entry()
{
  int offset = Seq_current_seq * Seq_seq_tags + 2;
  int location = Seq_current_entries;
  int i;
  int loc;
  int entry = -1;

  for (i = 0; i < Seq_current_entries; ++i)
    if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1))
      if (tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1) >= 0)  // Not deleted
        if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+2))
        {
          loc = tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+2);
          if (loc < location)
          {
            location = loc;
            entry = i;
          }
        }

  if (location == Seq_current_entries) return -1;

  return entry;
}

//---------------------------------------------------------------------------------------------

// Finds the entry with ordering number location. Returns -1 if not found.

int Seq_find_entry(int location)
{
  int offset = Seq_current_seq * Seq_seq_tags + 2;
  int i;

  for (i = 0; i < Seq_current_entries; ++i)
    if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1))
      if (tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1) >= 0)  // Not deleted
        if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+2))
          if (location == tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+2)) break;

  if (i >= Seq_current_entries) return -1;

  return i;
}

//---------------------------------------------------------------------------------------------

// Find a free sequence. Returns -1 if unable to locate an empty or create a new sequence.

int Seq_find_free()
{
  int seq;
  int offset;
  int entries;

  for (seq = 0; seq < Seq_number_of_seqs; ++seq)
  {
    offset = seq * Seq_seq_tags + 2;
    if (tagexists(Seq_current_file, Seq_tag+offset+1, 3))
    {
      if ((entries = tagread(Seq_current_file, Seq_tag+offset+1, 3)) <= 0)
      {
        break;    // Found an empty one
      }
      else
      {
        // Check if all the entries in the sequence are deleted.

        int i;
        int ndeleted;
        for (i = 0; i < entries; ++i)
        {
          if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1))
          {
            if (tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1) < 0)
              ++ndeleted;
          }
          else
            ++ndeleted;
        }

        if (ndeleted == entries) break;   // Found an empty sequence
      }
    }
  }

  if (seq < Seq_number_of_seqs) return seq;

  // There is no existing empty sequence, try for a new one.

  if (Seq_number_of_seqs < Seq_max_seq)
  {
    seq = Seq_number_of_seqs;
    ++Seq_number_of_seqs;
    return seq;
  }

  return -1;
}

//---------------------------------------------------------------------------------------------

// Finds the first active sequence. Returns -1 if there is none.

int Seq_find_first_seq()
{
  int seq = 0;
  int found = 0;
  int offset;
  int entries;

  if (Seq_number_of_seqs <= 0) return -1;

  while (!found && seq < Seq_number_of_seqs)
  {
    offset = seq * Seq_seq_tags + 2;
    if (tagexists(Seq_current_file, Seq_tag+offset+1, 3))
    {
      entries = tagread(Seq_current_file, Seq_tag+offset+1, 3);
      if (entries > 0)
      {
        // Check that all of the entries in the sequence are not deleted.

        int i;
        int ndeleted = 0;
        for (i = 0; i < entries; ++i)
        {
          if (tagexists(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1))
          {
            if (tagread(Seq_current_file, Seq_tag+offset+2, i*Seq_anchor_tags+1) < 0)
              ++ndeleted;
          }
          else
            ++ndeleted;
        }

        if (ndeleted != entries) found = 1;   // Found a non-empty sequence
      }
    }
    ++seq;
  }

  if (found) return --seq;

  return -1;
}

//---------------------------------------------------------------------------------------------

// Sets the current options from the dialogue box for the current sequence.

void Seq_set_options (int winhandle)
{
  // Sequence control stuff.

  string start = readicon(winhandle, 18);
  Seq_Start = stoi(start);                         // Specified starting number
  Seq_Name = readicon(winhandle, 16);              // Sequence name
}

//---------------------------------------------------------------------------------------------

// Updates the current sequence entry in the window.

void Seq_update_sequence()
{
  int offset = Seq_current_seq * Seq_seq_tags + 2;

  writeicon(Seq_win, 2, itos(Seq_current_seq));
  writeicon(Seq_win, 16, Seq_Name);
  writeicon(Seq_win, 18, itos(Seq_Start));

  if (Seq_CurrentNumberFormat == 2 || Seq_CurrentNumberFormat < 0)
  {
    select_icon (Seq_win, 5, 1);
    select_icon (Seq_win, 6, 0);
    select_icon (Seq_win, 7, 0);
  }
  else if (Seq_CurrentNumberFormat == 1)
  {
    select_icon (Seq_win, 5, 0);
    select_icon (Seq_win, 6, 1);
    select_icon (Seq_win, 7, 0);
  }
  else
  {
    select_icon (Seq_win, 5, 0);
    select_icon (Seq_win, 6, 0);
    select_icon (Seq_win, 7, 1);
  }

  if (Seq_current_entry >= 0)
  {
    if (tagexists(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags))
    {
      Seq_current_entry_value = tagread(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags);
      if (Seq_CurrentNumberFormat == 2)
        writeicon(Seq_win, 21, itos(Seq_current_entry_value));
      else if (Seq_CurrentNumberFormat == 1)
        writeicon(Seq_win, 21, -itors(Seq_current_entry_value));
      else
        writeicon(Seq_win, 21, itors(Seq_current_entry_value));
    }
    else
    {
      writeicon(Seq_win, 21, "");
      Seq_current_entry_value = 0;
    }
    if (tagexists(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+1))
    {
      Seq_current_references = tagread(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+1);
    }
    else
    {
      Seq_current_references = 0;
    }
    if (tagexists(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+3))
    {
      Seq_max_reference = tagread(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+3);
    }
    else
    {
      Seq_max_reference = 0;
    }
    set_icon_state(Seq_win, 11, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 12, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 15, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 22, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 23, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 24, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 14, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 29, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 30, 0x00000000, 0x00400000);
    if (Seq_current_references > 0)
      set_icon_state(Seq_win, 13, 0x00000000, 0x00400000);
    else
      set_icon_state(Seq_win, 13, 0x00400000, 0x00400000);
  }
  else
  {
    writeicon(Seq_win, 21, "");
    Seq_current_entry_value = 0;
    Seq_current_references = 0;
    Seq_max_reference = 0;
    set_icon_state(Seq_win, 11, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 12, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 15, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 22, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 23, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 24, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 14, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 29, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 13, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 30, 0x00400000, 0x00400000);
  }

  Seq_current_ref = 0;
}

//---------------------------------------------------------------------------------------------

// Load the default options.

void Seq_load_defaults()
{
  Seq_number_of_seqs = 0;
  Seq_current_seq = -1;
  Seq_current_entries = 0;
  Seq_max_entry_value = 0;
  Seq_NumberFormat = 2;       // Arabic
  Seq_CurrentNumberFormat = -1;
  Seq_Start = 1;
  Seq_current_ref = 0;
  Seq_current_references = 0;
  Seq_max_reference = 0;
  Seq_current_entry_value = 0;
  Seq_current_entry = -1;
}

//---------------------------------------------------------------------------------------------

// Save the current options with the document.

void Seq_save_options()
{
  int offset = Seq_current_seq * Seq_seq_tags + 2;

  tagwrite (Seq_current_file, Seq_tag+0, Seq_number_of_seqs, 0);
  tagwrite (Seq_current_file, Seq_tag+1, Seq_Version, 0);
  tagwrites (Seq_current_file, Seq_tag+offset, Seq_Name);
  tagwrite (Seq_current_file, Seq_tag+offset+1, Seq_Start, 0);
  tagwrite (Seq_current_file, Seq_tag+offset+1, Seq_max_entry_value, 1);
  tagwrite (Seq_current_file, Seq_tag+offset+1, Seq_NumberFormat, 2);
  tagwrite (Seq_current_file, Seq_tag+offset+1, Seq_current_entries, 3);
  // Seq_CurrentNumberFormat is written as it is changed to index 4.
}

//---------------------------------------------------------------------------------------------

// Load the current options from the document for a given sequence.

void Seq_load_options(int seq)
{
  int offset = seq * Seq_seq_tags + 2;

  if (seq < Seq_number_of_seqs && seq >= 0)
  {
    if (tagexists(Seq_current_file, Seq_tag+offset, 0))
      tagreads (Seq_current_file, Seq_tag+offset, Seq_Name);
    else
      Seq_Name = "";
    if (tagexists(Seq_current_file, Seq_tag+offset+1, 0))
      Seq_Start = tagread(Seq_current_file, Seq_tag+offset+1, 0);
    else
      Seq_Start = 1;
    if (tagexists(Seq_current_file, Seq_tag+offset+1, 1))
      Seq_max_entry_value = tagread(Seq_current_file, Seq_tag+offset+1, 1);
    else
      Seq_max_entry_value = Seq_Start - 1;
    if (tagexists(Seq_current_file, Seq_tag+offset+1, 2))
      Seq_NumberFormat = tagread(Seq_current_file, Seq_tag+offset+1, 2);
    else
      Seq_NumberFormat = 2;
    if (tagexists(Seq_current_file, Seq_tag+offset+1, 4))
      Seq_CurrentNumberFormat = tagread(Seq_current_file, Seq_tag+offset+1, 4);
    else
      Seq_CurrentNumberFormat = Seq_NumberFormat;
    if (tagexists(Seq_current_file, Seq_tag+offset+1, 3))
    {
      Seq_current_entries = tagread(Seq_current_file, Seq_tag+offset+1, 3);
      if (Seq_current_entries > 0)
      {
        Seq_current_entry = Seq_first_entry(); // Find earliest non-deleted entry
        if (Seq_current_entry >= 0)
        {
          if ((tagexists(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags)))
          {
            Seq_current_entry_value = tagread(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags);
          }
          else
            Seq_current_entry_value = 0;
          if ((tagexists(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+1)))
            Seq_current_references = tagread(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+1);
          else
            Seq_current_references = 0;
          if ((tagexists(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+3)))
            Seq_max_reference = tagread(Seq_current_file, Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+3);
          else
            Seq_max_reference = 0;
        }
        else
        {
          Seq_current_entry_value = 0;
          Seq_current_references = 0;
          Seq_max_reference = 0;
        }
      }
      else
      {
        Seq_current_entry = -1;
        Seq_current_entry_value = 0;
        Seq_current_references = 0;
        Seq_max_reference = 0;
      }
    }
    else
    {
      Seq_current_entries = 0;
      Seq_current_entry = -1;
      Seq_current_entry_value = 0;
      Seq_current_references = 0;
      Seq_max_reference = 0;
    }
  }
  else
  {
      Seq_Name = "";
      Seq_Start = 1;
      Seq_max_entry_value = 0;
      Seq_NumberFormat = 2;
      Seq_CurrentNumberFormat = Seq_NumberFormat;
      Seq_current_entries = 0;
      Seq_current_entry = -1;
      Seq_current_entry_value = 0;
      Seq_current_references = 0;
      Seq_max_reference = 0;
  }

  Seq_current_ref = 0;
}

//---------------------------------------------------------------------------------------------

// Loads the saved options from the document. Returns true always.

int Seq_load_document()
{
  Seq_load_defaults();			// Set default values initially

  if (tagexists(Seq_current_file, Seq_tag+0, 0))
  {
    Seq_number_of_seqs = tagread (Seq_current_file, Seq_tag+0, 0);

    // Find the first active sequence.

    Seq_current_seq = Seq_find_first_seq();
  }
  else
    Seq_current_seq = -1;

  Seq_load_options(Seq_current_seq);        // Load options for sequence

  return 1;
}

//---------------------------------------------------------------------------------------------

// Fills in the ordering information for the anchor located at bookmark bm and
// for anchor number anchor.

void Seq_set_order(int bm, int anchor)
{
  // We must scan through the existing anchors in the current sequence to determine
  // our location relative to those. We go through these in their document order.

  string bookmark = "SeqA_" + itos(Seq_current_seq) + "_";
  int bm1;
  int offset = Seq_current_seq * Seq_seq_tags + 2;
  int entry = Seq_first_entry();
  int location = -1;
  int old_entry;
  int chapter = bmchapter(bm);

  if (entry >= 0)
  {
    // Get the bookmark and ordering for the current entry.

    bm1 = bmfind(bookmark + itos(entry));
    location = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+2);

    // Search for a bookmark that is beyond our current location.

    while (entry >= 0 && bm1 > 0 && bmcmp(bm1, bm) <= 0 && bmchapter(bm1) <= chapter)
    {
      // Try the next anchor.
      
      entry = Seq_next_entry(entry);
      if (entry >= 0)
      {
        bm1 = bmfind(bookmark + itos(entry));
        location = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+2);
      }
    }
  }
  
  // Did we find one?

  if (entry >= 0)
  {
    // Yes, location is the ordering number that we want for our new anchor.

    tagwrite(Seq_current_file, Seq_tag+offset+2, location, anchor*Seq_anchor_tags+2);

    // Now update all of the remaining higher ordering numbers.

    while (entry >= 0)
    {
      location++;
      old_entry = entry;
      entry = Seq_next_entry(entry);
      tagwrite(Seq_current_file, Seq_tag+offset+2, location, old_entry*Seq_anchor_tags+2);
    }
  }
  else
  {
    // We did not locate any anchor beyond our new one, so add it to the end.

    ++location;
    tagwrite(Seq_current_file, Seq_tag+offset+2, location, anchor*Seq_anchor_tags+2);
  }
}

//---------------------------------------------------------------------------------------------

// Deletes the anchor anchor at the bookmark bm for sequence seq. refs is the number of references to
// the anchor. If bm is zero or less on entry, then it is assumed that the anchor value in the document
// has already been deleted, and so its entry in the sequence is just marked as being deleted. If keep
// is true on entry then the text associated with the anchor is not removed.

void Seq_delete_anchor(int bm, int seq, int seq_entries, int anchor, int anchor_value, int refs, int keep)
{
  string svalue;
  int offset = seq * Seq_seq_tags + 2;
  int entry;
  int format;
  int location;
  int loc;
  int bm1;
  int bm2;
  int current_max;

  if (refs <= 0) // Only delete if there are no references
  {
    // Does the anchor's bookmark exist?

    if (bm > 0)
    {
      // It does, remove it and the associated text from the document.

      if (!keep)
      {
        bm1 = bmcreate("Seq_temp3");
        bm2 = bmcreate("Seq_temp4");

        if (tagexists(Seq_current_file, Seq_tag+offset+1, 4))
          format = tagread(Seq_current_file, Seq_tag+offset+1, 4); // Existing number format
        else
          format = Seq_NumberFormat;

        if (format == 2)
          svalue = itos(anchor_value);
        else if (format == 1)
          svalue = -itors(anchor_value);
        else
          svalue = itors(anchor_value);

        setbmtobm(bm1, bm);
        if (bmsearch(bm1, bm2, svalue, 0) == 0)
        {
          if (bmcmp(bm,bm1) == 0)
          {
            setcarettobm(bm1);
            setzonetobm(bm2);
            if (candelete()) deleteselection();
          }
        }

        bmdelete(bm1);
        bmdelete(bm2);
      }

      bmdelete(bm);  // Delete the anchor's bookmark
    }

    // If we are dealing with the current sequence and anchor then obtain the
    // next entry to display in the control window.

    if (seq == Seq_current_seq && anchor == Seq_current_entry)
    {
      entry = Seq_previous_entry(Seq_current_entry);
      if (entry < 0) entry = Seq_next_entry(Seq_current_entry);
      Seq_current_entry = entry;
      Seq_set_options(Seq_win);
      Seq_update_sequence();
    }

    // Mark the anchor as being deleted.

    if (tagexists(Seq_current_file, Seq_tag+offset+2, anchor*Seq_anchor_tags+1))
      tagwrite(Seq_current_file, Seq_tag+offset+2, -1, anchor*Seq_anchor_tags+1);

    // Adjust the sequence numbers in the later anchors,

    if (tagexists(Seq_current_file, Seq_tag+offset+2, anchor*Seq_anchor_tags+2))
    {
      location = tagread(Seq_current_file, Seq_tag+offset+2, anchor*Seq_anchor_tags+2);
      for (entry = 0; entry < seq_entries; ++entry)
        if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1))
          if (tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1) >= 0)     // Not deleted
            if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+2))
            {
              loc = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+2);
              if (loc > location)
              {
                --loc;
                tagwrite(Seq_current_file, Seq_tag+offset+2, loc, entry*Seq_anchor_tags+2);
              }
            }
    }

    // Ajust the maximum value in the sequence if the anchor had that value.

    if (tagexists(Seq_current_file, Seq_tag+offset+1, 1))
    {
      current_max = tagread(Seq_current_file, Seq_tag+offset+1, 1);
      if (anchor_value >= current_max)
      {
        --current_max;
        tagwrite(Seq_current_file, Seq_tag+offset+1, current_max, 1);

        if (seq == Seq_current_seq)
          Seq_max_entry_value = current_max;
      }
    }
  }
  else
  {
    string s = "{SEQ_03}";
    string s1 = "{SEQ_04}";
    translate(s);
    translate(s1);
    messagebox(s + itos(refs) + s1);
  }
}

//---------------------------------------------------------------------------------------------

// Deletes the current reference at the bookmark bm for sequence seq, anchor anchor and
// reference ref.
// If bm is zero or less on entry, then it is assumed that the reference value in the document
// has already been deleted, and so we just decrement the reference count. The new reference count
// for the anchor is returned. If keep is true on entry then the text associated with the anchor is
// not removed.

int Seq_delete_reference(int bm, int seq, int anchor, int anchor_value, int ref, int keep)
{
  string svalue;
  int offset = seq * Seq_seq_tags + 2;
  int bm1;
  int bm2;
  int count = 0;
  int format;

  // Does the reference exist in the document?

  if (bm > 0)
  {
    // It does, remove its text and bookmark.

    if (!keep)
    {
      bm1 = bmcreate("Seq_temp5");
      bm2 = bmcreate("Seq_temp6");

      if (tagexists(Seq_current_file, Seq_tag+offset+1, 4))
        format = tagread(Seq_current_file, Seq_tag+offset+1, 4); // Existing number format
      else
        format = Seq_NumberFormat;

      if (format == 2)
        svalue = itos(anchor_value);
      else if (format == 1)
        svalue = -itors(anchor_value);
      else
        svalue = itors(anchor_value);

      setbmtobm(bm1, bm);

      if (bmsearch(bm1, bm2, svalue, 0) == 0)
      {
        if (bmcmp(bm,bm1) == 0)
        {
          setcarettobm(bm1);
          setzonetobm(bm2);
          if (candelete()) deleteselection();
        }
      }

      bmdelete(bm1);
      bmdelete(bm2);
    }

    bmdelete(bm);  // Delete the reference's bookmark.
  }

  // Decrement the reference count for the anchor.

  if (tagexists(Seq_current_file, Seq_tag+offset+2, anchor*Seq_anchor_tags+1))
  {
    count = tagread(Seq_current_file, Seq_tag+offset+2, anchor*Seq_anchor_tags+1);
    --count;
    tagwrite(Seq_current_file, Seq_tag+offset+2, count, anchor*Seq_anchor_tags+1);
  }

  // Adjust the maximum reference number if needed.

  if (count <= 0)
  {
    tagwrite(Seq_current_file, Seq_tag+offset+2, 0, anchor*Seq_anchor_tags+3);
  }
  else if (tagexists(Seq_current_file, Seq_tag+offset+2, anchor*Seq_anchor_tags+3))
  {
    if (ref >= tagread(Seq_current_file, Seq_tag+offset+2, anchor*Seq_anchor_tags+3))
      tagwrite(Seq_current_file, Seq_tag+offset+2, ref, anchor*Seq_anchor_tags+3);
  }

  // If we are dealing with the current sequence and anchor then update the display.

  if (seq == Seq_current_seq && anchor == Seq_current_entry)
  {
    Seq_set_options(Seq_win);
    Seq_update_sequence();
  }

  return count;
}

//---------------------------------------------------------------------------------------------

// Renumbers the entries for the current sequence. If ordered is true the entries
// are renumbered according to their order as they occur in the document.
// This function also attempts to correct for any bookmarks/entries that have
// been accidentally deleted by the user.

void Seq_renumber(int ordered)
{
  int entry = 0;
  int ref;
  int refs;
  int offset = Seq_current_seq * Seq_seq_tags + 2;
  int value;
  int newval = -1;
  int format;
  int bm;
  int bm1;
  int bm2;
  int len;
  int old_len;
  int i;
  int count = 0;
  string bookmark;
  string svalue;
  string snewval;

  if (Seq_current_entries > 0)
  {
//    swi13(0x406c0, 0, 0, 0);

    bm = bmcreate("Seq_temp");
    bm1 = bmcreate("Seq_temp1");

    if (tagexists(Seq_current_file, Seq_tag+offset+1, 4))
      format = tagread(Seq_current_file, Seq_tag+offset+1, 4);   // Existing number format

    if (ordered) entry = Seq_first_entry();

    while (entry < Seq_current_entries && entry >= 0)
    {
      if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1))
      {
        if (tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1) >= 0) // If entry is not deleted
        {
          if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags))
          {
            value = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags);
            if (format == 2)
              svalue = itos(value);
            else if (format == 1)
              svalue = -itors(value);
            else
              svalue = itors(value);

            newval = Seq_Start + count;
            tagwrite(Seq_current_file, Seq_tag+offset+2, newval, entry*Seq_anchor_tags);
            if (Seq_NumberFormat == 2)
              snewval = itos(newval);
            else if (Seq_NumberFormat == 1)
              snewval = -itors(newval);
            else
              snewval = itors(newval);

            len = slen(snewval);
            old_len = slen(svalue);

            // Obtain the number of references to the anchor.

            if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1))
              refs = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1);
            else
              refs = 0;

            // Do the anchor first.

            bookmark = "SeqA_" + itos(Seq_current_seq) + "_" + itos(entry);
            if (bm2 = bmfind(bookmark))
            {
              setbmtobm(bm, bm2);
              if (bmsearch(bm, bm1, svalue, 0) == 0)
              {
                if (bmcmp(bm,bm2) == 0)
                {
                  setcarettobm(bm2);
//                  setzonetobm(bm1);
//                  if (candelete())
//                  {
                    i = old_len;
                    while (i-- > 0) type("{DeleteF}"); // Note, goes a lot faster than overtyping zone
                    type(snewval);
//                    bm2 = bmcreate(bookmark);  // Replace the bookmarks (only needed if overtyping zone and using RJETs bookmark manager)
//                    setbmtocaret(bm2);
                    i = len;
                    while (i-- > 0) bmmove(bm2, 0, 0);
//                    bm = bmcreate("Seq_temp");

                    // Update its ordering information if we are reordering.

                    if (ordered)
                      tagwrite(Seq_current_file, Seq_tag+offset+2, count, entry*Seq_anchor_tags+2);
//                  }
                }
                else
                {
                  // It looks like the text has been deleted by the user, just insert the new value.

                  setcarettobm(bm2);
                  type(snewval);
                  i = len;
                  while (i-- > 0) bmmove(bm2, 0, 0);
                }
              }
              else
              {
                // It looks like the text has been deleted by the user, just insert the new value.

                setcarettobm(bm2);
                type(snewval);
                i = len;
                while (i-- > 0) bmmove(bm2, 0, 0);
              }

              ++count;

              // Now do the references.

              bookmark = "SeqR_" + itos(Seq_current_seq) + "_" + itos(entry) + "_";

              for (ref = 0; ref < refs; ++ref)
              {
                if (bm2 = bmfind(bookmark + itos(ref)))
                {
                  setbmtobm(bm,bm2);
                  if (bmsearch(bm, bm1, svalue, 0) == 0)
                  {
                    if (bmcmp(bm,bm2) == 0)
                    {
                      setcarettobm(bm2);
//                      setzonetobm(bm1);
//                      if (candelete())
//                      {
                        i = old_len;
                        while (i-- > 0) type("{DeleteF}"); // Note, goes a lot faster than overtyping zone
                        type(snewval);
//                        bm2 = bmcreate(bookmark + itos(ref));  // Replace the bookmarks (only needed if overtyping zone and using RJETs bookmark manager)
//                        setbmtocaret(bm2);
                        i = len;
                        while (i-- > 0) bmmove(bm2, 0, 0);
//                        bm = bmcreate("Seq_temp");
//                      }
                    }
                    else
                    {
                      // It looks like the text has been deleted by the user, just insert
                      // the new value.

                      setcarettobm(bm2);
                      type(snewval);
                      i = len;
                      while (i-- > 0) bmmove(bm2, 0, 0);
                    }
                  }
                  else
                  {
                    // It looks like the text has been deleted by the user, just insert
                    // the new value.

                    setcarettobm(bm2);
                    type(snewval);
                    i = len;
                    while (i-- > 0) bmmove(bm2, 0, 0);
                  }
                }
                else
                {
                  // We did not locate the bookmark for the reference. Assume that the user has
                  // deleted it. In this case we just finish it off by deleting it properly.

                  i = Seq_delete_reference(0, Seq_current_seq, entry, value, ref, 0);
                  if (entry == Seq_current_entry)
                  {
                    Seq_current_references = i;
                    if (tagexists(Seq_current_file,  Seq_tag+offset+2, entry*Seq_anchor_tags+3))
                      Seq_max_reference = tagread(Seq_current_file,  Seq_tag+offset+2, entry*Seq_anchor_tags+3);
                  }
                }
              }
            }
            else
            {
              // We did not find the anchor, assume that the user has deleted it.
              // In this case we finish it off by deleting it properly.

              // First delete the references to this anchor.

              bookmark = "SeqR_" + itos(Seq_current_seq) + "_" + itos(entry) + "_";

              for (ref = 0; ref < refs; ++ref)
              {
                bm2 = bmfind(bookmark + itos(ref));
                Seq_delete_reference(bm2, Seq_current_seq, entry, value, ref, 0);
              }

              // Now the anchor itself.

              Seq_delete_anchor(0, Seq_current_seq, Seq_current_entries, entry, value, 0, 0);

              if (entry == Seq_current_entry)
              {
                Seq_current_references = 0;
                Seq_max_reference = 0;
              }
            }
          }
        }
      }

      if (ordered)
        entry = Seq_next_entry(entry);
      else
        ++entry;
    }

    bmdelete(bm);
    bmdelete(bm1);

    // If we are re-ordering then we can update the current maximum anchor value in the
    // sequence.

    if (ordered) Seq_max_entry_value = newval;

    tagwrite(Seq_current_file, Seq_tag+offset+1, Seq_NumberFormat, 4);  // New format for sequence
    Seq_CurrentNumberFormat = Seq_NumberFormat;
    Seq_update_sequence();
    Seq_save_options();
//    swi13(0x406c1, 0, 0, 0);
  }
}

//---------------------------------------------------------------------------------------------

// Deletes a sequence. The values in the text are removed as well if keep is not
// true on entry.
// If bms is > 0 on entry then bms and bme are taken as bookmarks indicating the start and
// end of the region to be deleted. Only anchors and references in this region are removed.

void Seq_delete(int sequence, int bms, int bme, int keep)
{
  int entry = 0;
  int nentries = 0;
  int ref;
  int refs;
  int offset = sequence * Seq_seq_tags + 2;
  int value;
  int bm;
  int bma;
  int do_delete;
  int ref_count;
  int max_ref;
  string bookmark;

  if (tagexists(Seq_current_file, Seq_tag+offset+1, 3))
    nentries = tagread(Seq_current_file, Seq_tag+offset+1, 3);  // Number of entries in the sequence

  if (nentries > 0)
  {
    while (entry < nentries)
    {
      if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1))
      {
        if (tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1) >= 0) // If entry is not deleted
        {
          if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags))
          {
            value = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags);

            // Locate the anchor.

            bookmark = "SeqA_" + itos(sequence) + "_" + itos(entry);
            bma = bmfind(bookmark);

            if (bms > 0)
            {
              // Check if the anchor is in the range of our bookmarks.

              do_delete = bma > 0 && bmcmp(bma, bms) >= 0 && bmcmp(bma, bme) <= 0;
            }
            else
              do_delete = 1;

            // Obtain the number of references to the anchor.

            ref_count = 0;
            max_ref = 0;
            if (tagexists(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1))
            {
              refs = tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags+1);

              if (refs > 0)
              {
                // Delete the references.

                bookmark = "SeqR_" + itos(sequence) + "_" + itos(entry) + "_";

                for (ref = 0; ref < refs; ++ref)
                {
                  // Find and destroy the current reference.

                  bm = bmfind(bookmark + itos(ref));
                  if (bm > 0) max_ref = ref;

                  // Check if the reference is in the range of our bookmarks if we
                  // need to.

                  if (do_delete)
                  {
                    // We are deleting the anchor so we need to delete its references.

                    Seq_delete_reference(bm, sequence, entry, value, ref, keep);
                  }
                  else
                  {
                    // We are not deleting the anchor, so check if the reference has to be
                    // deleted.

                    if (bm > 0 && bmcmp(bm, bms) >= 0 && bmcmp(bm, bme) <= 0)
                      ref_count = Seq_delete_reference(bm, sequence, entry, value, ref, keep);
                  }
                }
              }
            }

            // Now delete the anchor if required.

            if (do_delete) Seq_delete_anchor(bma, sequence, Seq_current_entries, entry, value, 0, keep);

            if (sequence == Seq_current_seq && entry == Seq_current_entry)
            {
              Seq_current_references = ref_count;
              Seq_max_reference = max_ref;
            }
          }
        }
      }

      ++entry;
    }

    if (sequence == Seq_current_seq)
    {
      Seq_update_sequence();
      Seq_save_options();
    }

  }
}

//---------------------------------------------------------------------------------------------

// Checks if the region to be deleted specified by bm1 to bm2 contains any of our
// bookmarks, and if so tidies things up.

void Seq_check_delete(int bm1, int bm2)
{
  int seq;

  // Go through each sequence deleting any anchors or bookmarks in the
  // indicated region.

  for (seq = 0; seq < Seq_number_of_seqs; ++seq)
    Seq_delete(seq, bm1, bm2, 1);
}

//---------------------------------------------------------------------------------------------

// Called by 7!bm when a region is cut or deleted. The current zone
// must be set to the region.

int Seq_bmdeleter(int cut)
{
  int bm1;
  int bm2;

  // We need to check if any of our bookmarks are in the region.

  if (Seq_number_of_seqs > 0 && zoneorder() != 0)
  {
    bm1 = bmcreate("Seq_zonestart");
    bm2 = bmcreate("Seq_zoneend");
    setbmtozone(bm1, bm2);
    Seq_check_delete(bm1, bm2);
    bmdelete(bm1);
    bmdelete(bm2);
  }

  return 0;       // We don't claim the event
}

//---------------------------------------------------------------------------------------------

// Called by 7!bm when a delete key is pressed.

int Seq_keypress(int key)
{
  int bm1;
  int bm2;

  if (Seq_number_of_seqs > 0)
  {
    switch (key)
    {
      case 127:                    // Delete <
      case 6664:
        bm1 = bmcreate("Seq_keystart");
        bm2 = bmcreate("Seq_keyend");
        setbmtocaret(bm1);
        setbmtobm(bm2, bm1);
        bmmove(bm1, 0, 0);         // Previous character
        Seq_check_delete(bm1, bm2);
        bmdelete(bm1);
        bmdelete(bm2);
        break;

      case 395:                    // Delete >
        bm1 = bmcreate("Seq_keystart");
        bm2 = bmcreate("Seq_keyend");
        setbmtocaret(bm1);
        setbmtobm(bm2, bm1);
        bmmove(bm2, 1, 0);         // Next character
        Seq_check_delete(bm1, bm2);
        bmdelete(bm1);
        bmdelete(bm2);
        break;

      // Other cases are handled by 7!bm which sets the appropriate zone and
      // then calls Seq_bmdeleter.

/*
      case 2175:                   // Delete word before cursor
      case 2568:
        setbmtocaret(bm1);
        bmmove(bm1, 0, 5);         // Start of previous word
        setbmtobm(bm2, bm1);
        bmmove(bm2, 1, 1);         // End of that word
        Seq_check_delete(bm1, bm2);
        break;

      case 411:                    // Delete word at cursor
        setbmtocaret(bm1);
        setbmtobm(bm2, bm1);
        bmmove(bm1, 0, 1);         // Start of word
        bmmove(bm2, 1, 1);         // End of word
        Seq_check_delete(bm1, bm2);
        break;
*/

      default:
        break;
    }
  }

  return key;
}

//=============================================================================================

////////////////////////////////////////////////
// Sequence Window Code                       //
////////////////////////////////////////////////

// Display the sequences window.

int Seq_Window()
{
  if (Seq_current_seq >= 0)
  {
    writeicon(Seq_win, 2, itos(Seq_current_seq));
    writeicon(Seq_win, 16, Seq_Name);
    writeicon(Seq_win, 18, itos(Seq_Start));
    set_icon_state(Seq_win, 2, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 5, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 6, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 7, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 8, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 9, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 10, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 11, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 12, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 13, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 15, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 16, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 18, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 19, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 20, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 22, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 23, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 24, 0x00000000, 0x00400000);
  }
  else
  {
    writeicon(Seq_win, 2, "");
    writeicon(Seq_win, 16, "");
    writeicon(Seq_win, 18, "");
    set_icon_state(Seq_win, 2, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 5, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 6, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 7, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 8, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 9, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 10, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 11, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 12, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 13, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 14, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 15, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 16, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 18, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 19, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 20, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 22, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 23, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 24, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 29, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 30, 0x00400000, 0x00400000);
  }

  if (Seq_CurrentNumberFormat == 2)
  {
    select_icon (Seq_win, 5, 1);
    select_icon (Seq_win, 6, 0);
    select_icon (Seq_win, 7, 0);
  }
  else if (Seq_CurrentNumberFormat == 1)
  {
    select_icon (Seq_win, 5, 0);
    select_icon (Seq_win, 6, 1);
    select_icon (Seq_win, 7, 0);
  }
  else
  {
    select_icon (Seq_win, 5, 0);
    select_icon (Seq_win, 6, 0);
    select_icon (Seq_win, 7, 1);
  }

  if (Seq_current_entry >= 0)
  {
    // There is a current entry selected in the sequence.

    writeicon(Seq_win, 21, itos(Seq_current_entry_value));
    set_icon_state(Seq_win, 11, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 12, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 13, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 15, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 14, 0x00000000, 0x00400000);
    set_icon_state(Seq_win, 29, 0x00000000, 0x00400000);
  }
  else
  {
      // There is no current entry selected in the sequence.

    writeicon(Seq_win, 21, "");
    set_icon_state(Seq_win, 11, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 12, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 13, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 15, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 14, 0x00400000, 0x00400000);
    set_icon_state(Seq_win, 29, 0x00400000, 0x00400000);
  }

  display_window(Seq_win, 20, 1);

  return(1);
}

//---------------------------------------------------------------------------------------------

// Called when a sequence menu entry is opened. This assumes that ordering numbers
// increment uniformally from 0.

int Seq_menu_flags(int n, string& s)
{
  if (Seq_find_entry(n) == Seq_current_entry)
    return 1;
  else
    return 0;
}

//---------------------------------------------------------------------------------------------

// Called when a sequence menu entry is clicked. This assumes that ordering numbers
// increment uniformally from 0.

int Seq_menu_entry(int n, int subcode)
{
  Seq_current_entry = Seq_find_entry(n);
  Seq_update_sequence();

  return 1;
}

//---------------------------------------------------------------------------------------------

// Sets up the entry selection menu.

void Seq_set_menu(int seq)
{
  int entry;
  int offset = seq * Seq_seq_tags + 2;
  string name;
  int nentries = 0;

  // Get rid of any existing menu.

  if (Seq_menu)
  {
    delete_menu(Seq_menu);
    Seq_menu = 0;
  }

  // Locate the tag containing the entries for the current sequence.

  if (seq < Seq_number_of_seqs)
  {
    Seq_menu = create_menu("{SEQ_01}");
    entry = Seq_first_entry();
    while (entry >= 0)
    {
      if (Seq_CurrentNumberFormat == 2)
        name = itos(tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags));
      else if (Seq_CurrentNumberFormat == 1)
        name = -itors(tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags));
      else
        name = itors(tagread(Seq_current_file, Seq_tag+offset+2, entry*Seq_anchor_tags));
      addentry_menu(Seq_menu, "Seq_menu_entry", "Seq_menu_flags", "", "", name);

      entry = Seq_next_entry(entry);
      ++nentries;
    }

    if (nentries <= 0) // If no entries get rid of menu
    {
      delete_menu(Seq_menu);
      Seq_menu = 0;
    }
  }
}

//---------------------------------------------------------------------------------------------

// Opens the menus off the buttons on the entry control window.

void Seq_openwinmenu (int winhandle, int icon)
{
  Seq_set_menu(Seq_current_seq);     // Make the menu
  if (Seq_menu) open_menu(Seq_menu); // Show the menu
}

//---------------------------------------------------------------------------------------------




void Seq_close(int handle)
{
 if(Seq_win)
 {
  close_window(Seq_win);
  delete_window(Seq_win);
  Seq_win=0;
  Seq_current_file=0;
  Seq_current_view=0;
 }
}




//  Mouse click event handler for the sections style window

void Seq_click (int winhandle, int icon, int bbits, int mx, int my)
{
  if (bbits == 2 && icon != 24) return;	// Return if not left or right mouse button (except for menu icon)

  switch(icon)
  {
    case 8:                   // Sequence bump down icon
      if (Seq_current_seq > 0)
      {
        Seq_set_options(winhandle);
        Seq_save_options();
        --Seq_current_seq;
        Seq_load_options(Seq_current_seq);
        Seq_update_sequence();
      }
      else
        bbc_vdu(7);

      break;

    case 9:                   // Sequence bump up icon
      if (Seq_current_seq < Seq_number_of_seqs-1)
      {
        Seq_set_options(winhandle);
        Seq_save_options();
        ++Seq_current_seq;
        Seq_load_options(Seq_current_seq);
        Seq_update_sequence();
      }
      else
        bbc_vdu(7);

      break;

    case 19:                   // Start number bump down icon
      --Seq_Start;
      writeicon(Seq_win, 18, itos(Seq_Start));
      break;

    case 20:                   // Start number bump up icon
      ++Seq_Start;
      writeicon(Seq_win, 18, itos(Seq_Start));
      break;

    case 5:                    // Arabic
      select_icon(winhandle, 5, 1);
      select_icon(winhandle, 6, 0);
      select_icon(winhandle, 7, 0);
      Seq_NumberFormat = 2;
      break;

    case 6:                    // Roman lower
      select_icon(winhandle, 6, 1);
      select_icon(winhandle, 5, 0);
      select_icon(winhandle, 7, 0);
      Seq_NumberFormat = 1;
      break;

    case 7:                    // Roman upper
      select_icon(winhandle, 7, 1);
      select_icon(winhandle, 5, 0);
      select_icon(winhandle, 6, 0);
      Seq_NumberFormat = 0;
      break;

    case 22:                   // Entry bump down icon.
    {
      int entry = Seq_previous_entry(Seq_current_entry);
      if (entry >= 0)
      {
        Seq_current_entry = entry;
        Seq_update_sequence();
      }
      else
        bbc_vdu(7);

      break;
    }

    case 23:                   // Entry bump up icon.
    {
      int entry = Seq_next_entry(Seq_current_entry);
      if (entry >= 0)
      {
        Seq_current_entry = entry;
        Seq_update_sequence();
      }
      else
        bbc_vdu(7);

      break;
    }

    case 24:                   // Sequence entry menu
      Seq_openwinmenu(winhandle, icon);
      break;

    case 10:                   // New anchor.
      // We insert the new anchor at the current cursor position. The value is taken from the
      // current sequence.

      if (caretchar() >= 0)
      {
        string bookmark;
        int offset = Seq_current_seq * Seq_seq_tags + 2;
        int bm;
        int len;
        string anchor;

        if (Seq_first_entry() < 0)
        {
          // First active entry in this sequence.

          Seq_max_entry_value = Seq_Start;

          // Freeze the current number format.

          tagwrite(Seq_current_file, Seq_tag+offset+1, Seq_NumberFormat, 4);
          Seq_CurrentNumberFormat = Seq_NumberFormat;

          Seq_set_options(winhandle);
        }
        else
        {
          // Not the first entry in the sequence.

          ++Seq_max_entry_value;
        }

        Seq_current_entry = Seq_current_entries;
        bookmark = "SeqA_" + itos(Seq_current_seq) + "_" + itos(Seq_current_entry);
        bm = bmcreate(bookmark);
        setbmtocaret(bm);
        Seq_set_order(bm, Seq_current_entry);   // Insert the ordering information
        ++Seq_current_entries;
        Seq_current_entry_value = Seq_max_entry_value;
        if (Seq_NumberFormat == 2)
          anchor = itos(Seq_max_entry_value);
        else if (Seq_NumberFormat == 1)
          anchor = -itors(Seq_max_entry_value);
        else
          anchor = itors(Seq_max_entry_value);
        type(anchor);
        len = slen(anchor);
        while (len-- > 0) bmmove(bm, 0, 0);   // Move bookmark to before anchor text


        tagwrite(Seq_current_file, Seq_tag+offset+2, Seq_max_entry_value, Seq_current_entry*Seq_anchor_tags);
        tagwrite(Seq_current_file, Seq_tag+offset+2, 0, Seq_current_entry*Seq_anchor_tags+1);       // Number of refs
        tagwrite(Seq_current_file, Seq_tag+offset+2, 0, Seq_current_entry*Seq_anchor_tags+3);       // Maximum reference number
        Seq_save_options();
        Seq_update_sequence();
      }
      else
        bbc_vdu(7);

      break;

    case 14:                   // Renumber the current sequence
      Seq_set_options(winhandle);
      Seq_renumber(0);
      break;

    case 29:                   // Reorder the current sequence
      Seq_set_options(winhandle);
      Seq_renumber(1);
      break;

    case 28:                   // Create a new sequence.
    {
      int seq;

      if (Seq_number_of_seqs > 0)
      {
        Seq_set_options(winhandle);      // Save current sequence parameters
        Seq_save_options();
      }

      // Get a new sequence number.

      seq = Seq_find_free();

      if (seq >= 0)
      {
        int offset = seq * Seq_seq_tags + 2;

        // Default values for new sequence.

        Seq_current_seq = seq;
        Seq_current_entries = 0;
        Seq_current_entry = -1;
        Seq_Start = 1;
        Seq_max_entry_value = 0;
        Seq_CurrentNumberFormat = -1;
        Seq_NumberFormat = 2;
        Seq_Name = "";
        Seq_current_ref = 0;
        Seq_update_sequence();
        Seq_save_options();
        tagwrite(Seq_current_file, Seq_tag+offset+1, -1, 4);   // No number format yet
        set_icon_state(Seq_win, 2, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 5, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 6, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 7, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 8, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 9, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 10, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 11, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 12, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 13, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 14, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 15, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 16, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 18, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 19, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 20, 0x00000000, 0x00400000);
        set_icon_state(Seq_win, 22, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 23, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 24, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 29, 0x00400000, 0x00400000);
        set_icon_state(Seq_win, 30, 0x00400000, 0x00400000);
      }
      else
      {
        string s = "{SEQ_02}";
        translate(s);
        s += itos(Seq_max_seq);
        messagebox(s);
      }

      break;
    }

    case 30:                   // Delete a sequence
//      swi13(0x406c0, 0, 0, 0);
      Seq_set_options(Seq_win);
      Seq_delete(Seq_current_seq, 0, 0, 0);
//      swi13(0x406c1, 0, 0, 0);
      break;

    case 11:                   // Insert reference to the current sequence entry.
      // We insert a reference to the current anchor at the current cursor position.

      if (caretchar() >= 0 && Seq_current_entry >= 0)
      {
        int offset = Seq_current_seq * Seq_seq_tags + 2;
        string bookmark;
        int bm;
        string reference;
        int len;

        bookmark = "SeqR_" + itos(Seq_current_seq) + "_" + itos(Seq_current_entry) + "_" + itos(Seq_max_reference);
        ++Seq_current_references;
        ++Seq_max_reference;
        bm = bmcreate(bookmark);
        setbmtocaret(bm);
        if (Seq_CurrentNumberFormat == 2)
          reference = itos(Seq_current_entry_value);
        else if (Seq_CurrentNumberFormat == 1)
          reference = -itors(Seq_current_entry_value);
        else
          reference = itors(Seq_current_entry_value);
        type(reference);
        len = slen(reference);
        while (len-- > 0) bmmove(bm, 0, 0);   // Move bookmark to before reference text
        tagwrite(Seq_current_file, Seq_tag+offset+2, Seq_current_references, Seq_current_entry*Seq_anchor_tags+1);
        tagwrite(Seq_current_file, Seq_tag+offset+2, Seq_max_reference, Seq_current_entry*Seq_anchor_tags+3);
        Seq_update_sequence();
      }
      else
        bbc_vdu(7);

      break;

    case 12:                   // Find the current anchor in the document.
      if (Seq_current_entry >= 0)
      {
        string bookmark = "SeqA_" + itos(Seq_current_seq) + "_" + itos(Seq_current_entry);
        int bm = bmfind(bookmark);
        if (bm)
        {
          setcarettobm(bm);
          bmview(bm);
        }
        else
          bbc_vdu(7);
      }
      else
        bbc_vdu(7);

      break;

    case 13:                   // Find the next reference to the current anchor
      if (Seq_current_entry >= 0 && Seq_current_references > 0 && Seq_current_ref < Seq_max_reference)
      {
        string bookmark = "SeqR_" + itos(Seq_current_seq) + "_" + itos(Seq_current_entry) + "_";
        int bm;
        int found = 0;
        while (!found && Seq_current_ref < Seq_max_reference)
        {
          bm = bmfind(bookmark + itos(Seq_current_ref));
          if (bm)
          {
            setcarettobm(bm);
            bmview(bm);
            found = 1;
          }
          ++Seq_current_ref;
        }
        if (!found)
        {
          bbc_vdu(7);
          Seq_current_ref = 0;
        }
      }
      else
      {
        bbc_vdu(7);
        Seq_current_ref = 0;
      }

      break;

    case 15:                   // Delete an anchor or reference, depending on where we are in the document.
    {
      // First we need to find our if we are next to an anchor or a reference.

      string bookmark;
      int bm;
      int bm1;
      int ref = 0;
      int offset = Seq_current_seq * Seq_seq_tags + 2;

      bm1 = bmcreate("Seq_temp");
      setbmtocaret(bm1);

      // Try the anchor first.

      bookmark = "SeqA_" + itos(Seq_current_seq) + "_" + itos(Seq_current_entry);
      if (bm = bmfind(bookmark))
        if (bmcmp(bm,bm1) == 0)
        {
          Seq_delete_anchor(bm, Seq_current_seq, Seq_current_entries, Seq_current_entry,
                            Seq_current_entry_value, Seq_current_references, 0);
          bmdelete(bm1);
          break;
        }

      // Try for a reference.

      if (Seq_current_references > 0)
      {
        bookmark = "SeqR_" + itos(Seq_current_seq) + "_" + itos(Seq_current_entry) + "_";
        while (ref < Seq_max_reference)
        {
          if (bm = bmfind(bookmark + itos(ref)))
            if (bmcmp(bm,bm1) == 0)
            {
              Seq_current_references = Seq_delete_reference(bm, Seq_current_seq, Seq_current_entry, Seq_current_entry_value, ref, 0);
              Seq_max_reference = tagread(Seq_current_file,  Seq_tag+offset+2, Seq_current_entry*Seq_anchor_tags+3);
              break;
            }

          ++ref;
        }
      }

      bmdelete(bm1);
      break;
    }

    case 25:                    // Default option, dismiss window, saving options.
      Seq_set_options(winhandle);
      Seq_save_options();
      if (bbits == 4) Seq_close(winhandle);
      break;


    default:                   // Click anywhere else - don't do anything
      break;

  }
}

//---------------------------------------------------------------------------------------------

// Key press event handler for Serquence window.

int Seq_key_press (int winhandle, int icon, int key)
{
  switch (key)
  {
    case 13:			// Return or Esc, simulate mouse click over Dismiss
    case 27:
      Seq_click (winhandle, 25, 4, 0, 0);
      break;

    default:
      break;
   }

   return key;
}

//=============================================================================================



////////////////////////////////////////////////
// Sequence Main Menu functions               //
////////////////////////////////////////////////

// Called when the Sequence menu entry is clicked.

int Seq_command (int n, int sub)
{

  Seq_close(Seq_win);


  // Check if the file has changed.

  if (Seq_current_file != currentfile())
  {
    Seq_current_file = currentfile();	// Get file handle
    Seq_load_document ();	        // Load options saved with document
  }

  Seq_current_view=currentview();

  // Create the window, assign to handle.

  Seq_win=create_windowparent("Sequencer",viewhandle(currentview()));


  // Add handlers for when the user clicks on the sections styles window or when
  // s(he) hits a key with it displayed.

  addwindowhandler(0, Seq_win, "Seq_click");
  addwindowhandler(1, Seq_win, "Seq_key_press");
  addwindowhandler(7, Seq_win, "Seq_close");
  addwindowhandler(2, Seq_win, "Seq_close");


  Seq_Window();				// Display the Sequence window

  setactive(Seq_win);
}

//=============================================================================================

////////////////////////////////////////////////
// Event handlers.			      //
////////////////////////////////////////////////

// Called when a file is closed.

void Seq_file_closing (int user, int file)
{
 if (file == Seq_current_file)
 {
  Seq_close(Seq_win);
 }
}

//---------------------------------------------------------------------------------------------

// Called when a view is made active.

void Seq_view_active (int user, int view, int gain)
{
 if(view==Seq_current_view)
 {
  if(!gain)
  {
   Seq_close(Seq_win);
  }
 }
}

//---------------------------------------------------------------------------------------------

// Called when a view is closed.

void Seq_view_closed (int user, int view)
{
 if(view == Seq_current_view)
 {
  Seq_close(Seq_win);
 }
}

//=============================================================================================

////////////////////////////////////////////////
// Main Functions                             //
////////////////////////////////////////////////

void main(void)
{
  string help = "{SEQ_05}";

  Seq_load_defaults ();	// Set the default options


  // Add sections menu to applets menu.

  script_menu_initialise();             // Initialise the menu
  script_menu_insert("Seq_command", "", "", "", "{SEQ_00}");


  // Add handlers for files and views being closed and views made active.

  addeventhandler (0x02, 0, "Seq_file_closing");
  addeventhandler (0x06, 0, "Seq_view_closed");
  addeventhandler (0x215, 0, "Seq_view_active");

  // Define a macro to launch us off.

  if (isdefmacro("sequencer") == -1)
  {
    translate(help);
    defmacro3(0,"sequencer", "{Seq_command(0,0)}", "sequencer", help, "Sequence");
  }

  // Declare deletion handlers so that we can tidy up our bookmarks in deleted text.
  // We can only do this if !!!!!!!bm is being used. In this case we also take advantage
  // of that applet to hide our bookmarks. We do this as a macro so that it does not
  // cause interpreter errors if the bm_ functions do not exist.

//  if (objectexists("<OvationPro$Applets>.!!!!!!!bm") == 2)
//  {
//     macro("{bm_hide(\"Seq\");bm_registerdelfn(\"Seq_bmdeleter\");bm_registerdelkeyfn(\"Seq_keypress\");}");
//  }
}
